#define MEMSLOT_ROM		0
#define MEMSLOT_SL0		1
#define MEMSLOT_SL1		2
#define MEMSLOT_RAM		3

#define FOTON_CPU_FREQ			(16000000 / 9)
#define FOTON_TICKS_FRAME		(FOTON_CPU_FREQ / 50)
#define FOTON_TICKS_PER_SAMPLE  (FOTON_CPU_FREQ / SAMPLE_RATE)

namespace EmuFoton
{
#ifdef PALETTE_R8G8B8
unsigned int palette888[256];
#endif
#ifdef PALETTE_R3G3B2
unsigned char palette332[256];
#endif

unsigned char* ram = 0;
unsigned char* rom = 0;

unsigned char mem_layout[4];
unsigned char mem_layout_reg;

unsigned char video_mode;
unsigned short int video_base_adr;
bool video_blank;
unsigned char ink_color;
unsigned char paper_color;

unsigned short int video_offset_text;
unsigned short int video_offset_char;
unsigned short int video_offset_gfx;
unsigned short int video_offset_col;

unsigned char keyboard_row[10];
unsigned short int keyboard_row_active;

unsigned char joystick;

unsigned char sound_pin_state;

int sound_out_acc;
int sound_out_val;
int sound_out_div;

unsigned char port_rd(unsigned char adr)
{
  switch (adr)
  {
    case 0x80:	//PPI1-A memory configuration
      return mem_layout_reg;

    case 0x81:	//PPI1-B keyobard data
      return keyboard_row[keyboard_row_active];

    case 0x82:	//PPI1-C keyboard row and sound
      return keyboard_row_active | sound_pin_state;

    case 0x8c:	//joystick 1 - xx21RLDU
      return joystick;

    case 0x8d:	//joystick 2, actually tape input
      return 0;
  }

  return 0;
}

void port_wr(unsigned char adr, unsigned char val)
{
  switch (adr)
  {
    case 0x80:	//PPI1-A memory configuration
      {
        mem_layout[0] = val & 0x03;
        mem_layout[1] = (val >> 2) & 0x03;
        mem_layout[2] = (val >> 4) & 0x03;
        mem_layout[3] = (val >> 6) & 0x03;
        mem_layout_reg = val;
      }
      break;

    case 0x82:	//PPI1-C sound,tape out,REG LED,keyboard column select
      {
        keyboard_row_active = val & 0x0f;

        val &= 0x80;

        if (val != sound_pin_state)
        {
          sound_pin_state = val;

          IOList::add(CPU8080::time, 0, sound_pin_state);
        }
      }
      break;

    case 0x83:	//PPI1 control word
      break;

    case 0x84:	//PPI2-A video mode
      {
        video_base_adr = (val & 0xc0) << 8;
        video_mode = val | 0x0f;
      }
      break;

    case 0x85://PPI2-B printed output
      break;

    case 0x86:	//PPI2-C - bit 4 forced blanking, bit 5 ALF LED, other bits unused
      {
        video_blank = val & 0x10 ? false : true;
      }
      break;

    case 0x87:	//PPI2 control word
      break;

    case 0x88:	//text mode and border color
      {
        ink_color = (val & 0x0f);
        paper_color = (val >> 4);	//border color for mode 1,2
      }
      break;

    case 0x90:	//text buffer offset
      {
        video_offset_text = (val << 10) & 0x3800;
      }
      break;

    case 0x91:	//font offset
      {
        video_offset_char = (val << 10) & 0x3800;
      }
      break;

    case 0x92:	//pixel data offset
      {
        video_offset_gfx = (val << 10) & 0x2000 ^ 0x2000;
      }
      break;

    case 0x93:	//attribute data offset
      {
        video_offset_col = (val << 10) & 0x2000 ^ 0x2000;
      }
      break;
  }
}

void mem_wr(unsigned short int adr, unsigned char val)
{
  switch (mem_layout[adr >> 14])
  {
    case MEMSLOT_RAM:
    case MEMSLOT_ROM:
      ram[adr] = val;
      break;

    case MEMSLOT_SL0:
    case MEMSLOT_SL1:
      break;
  }
}

unsigned char mem_rd(unsigned short int adr)
{
  switch (mem_layout[adr >> 14])
  {
    case MEMSLOT_RAM:
      return ram[adr];

    case MEMSLOT_ROM:
      return rom[adr & 0x3fff];

    case MEMSLOT_SL0:
    case MEMSLOT_SL1:
      return 0xff;
  }

  return 0;
}

void keyboard_set_state(unsigned short int key_code, bool state)
{
  if (state)
  {
    keyboard_row[key_code >> 8] |= (key_code & 0xff);
  }
  else
  {
    keyboard_row[key_code >> 8] &= ~(key_code & 0xff);
  }
}

int update(int pad_state, bool audio_underrun)
{
  joystick = 0;

  if (pad_state & PAD_UP)    joystick |= 0x01;	//KLAD uses joystick port
  if (pad_state & PAD_DOWN)  joystick |= 0x02;
  if (pad_state & PAD_LEFT)  joystick |= 0x04;
  if (pad_state & PAD_RIGHT) joystick |= 0x08;
  if (pad_state & PAD_COIN)   joystick |= 0x10;

  keyboard_set_state(0x840, joystick & 0x01 ? true : false);	//Piton and Tetris use keyboard instead
  keyboard_set_state(0x820, joystick & 0x02 ? true : false);
  keyboard_set_state(0x880, joystick & 0x04 ? true : false);
  keyboard_set_state(0x810, joystick & 0x08 ? true : false);
  keyboard_set_state(0x801, joystick & 0x10 ? false : true);	//Piton and Tetris have inverted coin line

  CPU8080::time = 0;

  CPU8080::interrupt(0x38);

  int t = FOTON_TICKS_FRAME;

  if (audio_underrun) t += FOTON_TICKS_FRAME / 10;

  CPU8080::run(t);

  IOList::add(CPU8080::time, 0, sound_pin_state); //mark end of frame
  
  return t;
}

void _render_text_40(unsigned char** raster)
{
  int ink = palette332[ink_color & 0x0f];
  int pap = palette332[paper_color & 0x0f];

  int y = (RASTER_H - 192) / 2;

  int ptr = video_base_adr + video_offset_text;

  for (int cy = 0; cy < 24; ++cy)
  {
    int x = (RASTER_W - 40 * 6) / 2;

    for (int cx = 0; cx < 40; ++cx)
    {
      int chr = (ram[ptr++] << 3) + video_base_adr + video_offset_char;

      for (int py = y; py < y + 8; ++py)
      {
        unsigned char pb = ram[chr++];

        raster[py][x + 0] = pb & 0x80 ? ink : pap;
        raster[py][x + 1] = pb & 0x40 ? ink : pap;
        raster[py][x + 2] = pb & 0x20 ? ink : pap;
        raster[py][x + 3] = pb & 0x10 ? ink : pap;
        raster[py][x + 4] = pb & 0x08 ? ink : pap;
        raster[py][x + 5] = pb & 0x04 ? ink : pap;
      }

      x += 6;
    }

    ptr += 24;
    y += 8;
  }
}

void _render_text_32(unsigned char** raster)
{
  int y = (RASTER_H - 192) / 2;
  int ptr = video_base_adr + video_offset_text;

  for (int cy = 0; cy < 24; ++cy)
  {
    int x = 0;

    for (int cx = 0; cx < 32; ++cx)
    {
      unsigned char code = ram[ptr++];
      unsigned char attr = ram[(code >> 3) + video_base_adr + video_offset_col];
      unsigned char ink = palette332[attr & 0x0f];
      unsigned char pap = palette332[attr >> 4];

      int chr = (code << 3) + video_base_adr + video_offset_char;

      for (int py = y; py < y + 8; ++py)
      {
        unsigned char pb = ram[chr++];

        raster[py][x + 0] = pb & 0x80 ? ink : pap;
        raster[py][x + 1] = pb & 0x40 ? ink : pap;
        raster[py][x + 2] = pb & 0x20 ? ink : pap;
        raster[py][x + 3] = pb & 0x10 ? ink : pap;
        raster[py][x + 4] = pb & 0x08 ? ink : pap;
        raster[py][x + 5] = pb & 0x04 ? ink : pap;
        raster[py][x + 6] = pb & 0x02 ? ink : pap;
        raster[py][x + 7] = pb & 0x01 ? ink : pap;
      }

      x += 8;
    }

    y += 8;
  }
}

void _render_graphics(unsigned char** raster)
{
  int y = (RASTER_H - 192) / 2;

  int ptr = video_base_adr + video_offset_char;

  for (int cy = 0; cy < 24; ++cy)
  {
    int x = 0;

    for (int cx = 0; cx < 32; ++cx)
    {
      int chr = (ram[ptr++] + ((cy >> 3) << 8)) << 3;

      int col = chr + video_offset_col;
      int gfx = chr + video_offset_gfx;

      for (int py = y; py < y + 8; ++py)
      {
        unsigned char pc = ram[col++];

        unsigned char ink = palette332[pc & 0x0f];
        unsigned char pap = palette332[pc >> 4];

        unsigned char pb = ram[gfx++];

        raster[py][x + 0] = pb & 0x80 ? ink : pap;
        raster[py][x + 1] = pb & 0x40 ? ink : pap;
        raster[py][x + 2] = pb & 0x20 ? ink : pap;
        raster[py][x + 3] = pb & 0x10 ? ink : pap;
        raster[py][x + 4] = pb & 0x08 ? ink : pap;
        raster[py][x + 5] = pb & 0x04 ? ink : pap;
        raster[py][x + 6] = pb & 0x02 ? ink : pap;
        raster[py][x + 7] = pb & 0x01 ? ink : pap;
      }

      x += 8;
    }

    y += 8;
  }
}

void render_video(unsigned char** raster)
{
  unsigned char pap = palette332[paper_color & 0x0f];

  for (int i = 0; i < (RASTER_H - 192) / 2; ++i)
  {
    for (int j = 0; j < RASTER_W; ++j)
    {
      raster[i][j] = pap;
      raster[RASTER_H - 1 - i][j] = pap;
    }
  }

  if (video_blank)
  {
    for (int i = 0; i < 192; ++i)
    {
      for (int j = 0; j < 256; ++j)
      {
        raster[i + (RASTER_H - 192) / 2][j] = pap;
      }
    }

    return;
  }

  if (video_mode & 0x10)
  {
    _render_graphics(raster);
  }
  else
  {
    if (video_mode & 0x20)
    {
      _render_text_40(raster);
    }
    else
    {
      _render_text_32(raster);
    }
  }
}

void render_audio(int t)
{
  int prev_t = 0;

  for (int i = 0; i < IOList::length(); ++i)
  {
    int new_t = IOList::entry[i].time;
    
    int diff = (new_t - prev_t);

	prev_t = new_t;

    sound_out_val = IOList::entry[i].val ? SOUND_VOLUME : 0;

    int samples = diff / FOTON_TICKS_PER_SAMPLE;

    for (int j = 0; j < samples; ++j)
    {
      audio_queue_add(sound_out_val);
    }
  }
}

void init(void)
{
  ram = global_ram1;  //65536
  rom = global_ram2;  //16384

  for (int i = 0; i < 4; ++i) mem_layout[i] = MEMSLOT_ROM;

  mem_layout_reg = 0;

  video_mode = 0;
  video_base_adr = 0;
  video_blank = true;
  ink_color = 0;
  paper_color = 0;

  video_offset_text = 0;
  video_offset_char = 0;
  video_offset_gfx = 0;
  video_offset_col = 0;

  for (int i = 0; i < 10; ++i) keyboard_row[i] = 0;

  keyboard_row_active = 0;
  joystick = 0;

  sound_pin_state = 0;

  sound_out_acc = 0;
  sound_out_val = 0;
  sound_out_div = 0;

  //make palette

  for (int i = 0; i < 16; ++i)
  {
    int r = 0;
    int g = 0;
    int b = 0;

    if (i & 1)
    {
      g = (i & 2) ? 255 : 0;
      b = (i & 4) ? 255 : 0;
      r = (i & 8) ? 255 : 0;
    }
    else
    {
      g = (i & 2) ? 128 : 0;
      b = (i & 4) ? 128 : 0;
      r = (i & 8) ? 128 : 0;
    }

#ifdef PALETTE_R8G8B8
    palette888[i] = (r << 16) | (g << 8) | b;
#endif
#ifdef PALETTE_R3G3B2
    palette332[i] = ((r >> 5) << 5) | ((g >> 5) << 2) | (b >> 6);
#endif
  }

  CPU8080::init(mem_rd, mem_wr, port_rd, port_wr);
}

void reset(void)
{
  CPU8080::reset();
}

void rom_load(const char* filename, unsigned short int adr)
{
  file_load(filename, &rom[adr]);
}
}
