#define SPR_ATTR_ENABLE		0x01
#define SPR_ATTR_FLIP_H		0x08
#define SPR_ATTR_FLIP_V		0x02

#define TIA88_CPU_FREQ      (15750000 / 9)
#define TIA88_TICKS_VSYNC   57
#define TIA88_TICKS_FRAME   (TIA88_CPU_FREQ / 50)

namespace EmuTIA88
{
const double g_v[] = { 1.2071, 0.9971, 0.9259, 0.7159, 0.4912, 0.2812, 0.2100, 0.0000 };
const double r_v[] = { 1.5937, 1.3125, 1.1562, 0.8750, 0.7187, 0.4375, 0.2812, 0.0000 };
const double b_v[] = { 1.3523, 0.8750, 0.4773, 0.0000 };

#ifdef PALETTE_R8G8B8
unsigned int palette888[256];
#endif
#ifdef PALETTE_R3G3B2
unsigned char palette332[256];
#endif

unsigned char* pmem = 0;		//0000..dfff program ROM, e000..ffff RAM
unsigned char* vrom = 0;		//0000..7fff sprite character ROM
unsigned char* vram = 0;		//video RAM
unsigned char* vram_cache = 0;	//video RAM cache, pre-converted VRAM contents to speed up drawing

unsigned char vpal[16];			//palette RAM

unsigned char spr_x[16];
unsigned char spr_y[16];
unsigned char spr_n[16];
unsigned char spr_a[16];

unsigned char pio_d0;
unsigned char pio_d1;
unsigned char pio_d2;

unsigned char vsync;

unsigned char vram_sel;
unsigned char vram_en;
unsigned char bitplane_mask;
bool vram_cache_update;

unsigned char scr_xoff;
unsigned char scr_yoff;

bool input_analog;
int input_analog_value;

unsigned char port_rd(unsigned char adr)
{
  switch (adr)
  {
    case 0xD0: return pio_d0;
    case 0xD1: return pio_d1;
    case 0xD2: return pio_d2 | vsync;
  }

  return 0;
}

void port_wr(unsigned char adr, unsigned char val)
{
  switch (adr & 0xf0)
  {
    case 0x40:
      spr_y[adr - 0x40] = val ^ 0xff;	//sprite Y
      return;

    case 0x50:
      spr_x[adr - 0x50] = val ^ 0xff;	//sprite X
      return;

    case 0x60:
      spr_n[adr - 0x60] = val ^ 0xff;	//sprite num
      return;

    case 0x70:
      spr_a[adr - 0x70] = val ^ 0xff;	//sprite attr
      return;

    case 0xa0:	//palette
      adr -= 0xa0;
      vpal[adr] = palette332[val];
      return;
  }

  switch (adr)	//everything else
  {
    case 0xbc://vertical BG offset
      scr_yoff = 0;// (256 - val) & 0xff;
      return;

    case 0xbd://horizontal BG offset
      scr_xoff = 0;// (256 - val) & 0xff;
      return;

    case 0xbe://BG bitplane control, video page select
      vram_sel = val;
      return;

    case 0xbf://BG bitplane control
      vram_en = val;
	  bitplane_mask = 0;
	  if (vram_en & 0x01) bitplane_mask |= 0x01;
	  if (vram_en & 0x04) bitplane_mask |= 0x02;
	  if (vram_en & 0x10) bitplane_mask |= 0x04;
	  if (vram_en & 0x40) bitplane_mask |= 0x08;
      return;

    case 0xc0://PIT 0 ch 0
    case 0xc1://PIT 0 ch 1
    case 0xc2://PIT 0 ch 2
    case 0xc3://PIT 0 control word

      SoundTIA::write(CPU8080::time, adr, val);

    case 0xd4://PIT 1 ch 0, the secondary PIT missing in various versions and generally is not used
    case 0xd5://PIT 1 ch 1
    case 0xd6://PIT 1 ch 2
    case 0xd7://PIT 1 control word
    case 0xda://PIT 1 gate control
    case 0xd2://coin register control
    case 0xd3://PPI control word
      return;
  }
}

void convert_tiles(void)
{
	for (int id = 0; id < 256; ++id)
	{
		int ptr = id << 3;

		for (int y = 0; y < 8; ++y)
		{
			int pmask = 0x80;
			
			for (int x = 0; x < 8; ++x)
			{
				int col = (vram[ptr + 0x0800] & pmask ? 1 : 0) | (vram[ptr + 0x1000] & pmask ? 2 : 0) | (vram[ptr + 0x1800] & pmask ? 4 : 0) | (vram[ptr + 0x2000] & pmask ? 8 : 0);

				pmask >>= 1;

				vram_cache[id * 64 + y * 8 + x] = col;
			}

			++ptr;
		}
	}
}

void mem_wr(unsigned short int adr, unsigned char val)
{
  if (adr >= 0xe000)
  {
    pmem[adr] = val;

    return;
  }

  if (adr >= 0xb000 && adr <= 0xb7ff)
  {
    //if(!(vram_sel&64))
    {
      unsigned short int voff = adr & 0x7ff;

      if (!(vram_sel & 1))  vram[voff + 0x0000] = val;
      if (!(vram_sel & 2))  vram[voff + 0x0800] = val;
      if (!(vram_sel & 4))  vram[voff + 0x1000] = val;
      if (!(vram_sel & 8))  vram[voff + 0x1800] = val;
      if (!(vram_sel & 16)) vram[voff + 0x2000] = val;

      if (~vram_sel & (2 | 4 | 8 | 16)) vram_cache_update = true;
    }

    return;
  }
}

unsigned char mem_rd(unsigned short int adr)
{
  return pmem[adr];
}

int update(int pad_state, bool audio_underrun)
{
  pio_d0 = 0;
  pio_d1 = 0;
  pio_d2 = 0;

  if (!input_analog)
  {
    if (pad_state & PAD_LEFT)  pio_d0 |= 0x20;
    if (pad_state & PAD_RIGHT) pio_d0 |= 0x02;
    if (pad_state & PAD_UP)    pio_d1 |= 0x02;
    if (pad_state & PAD_DOWN)	 pio_d1 |= 0x20;
  }
  else
  {
    if (pad_state & PAD_LEFT)  input_analog_value += 4;
    if (pad_state & PAD_RIGHT) input_analog_value -= 4;

    if (input_analog_value < 96)  input_analog_value = 96;
    if (input_analog_value > 208) input_analog_value = 208;

    pio_d0 = input_analog_value;
  }

  if (pad_state & PAD_TEST) pio_d1 |= 0x80;
  if (pad_state & PAD_COIN) pio_d2 |= 0x10;
  if (pad_state & PAD_A)	pio_d2 |= 0x40;
  if (pad_state & PAD_B)	pio_d2 |= 0x20;

  CPU8080::time = 0;

  vsync |= 0x80;

  CPU8080::run(TIA88_TICKS_VSYNC);

  vsync &= ~0x80;

  int t = TIA88_TICKS_FRAME - TIA88_TICKS_VSYNC;

  if (audio_underrun) t += TIA88_TICKS_FRAME / 10;

  CPU8080::run(t);

  SoundTIA::write(CPU8080::time, 0, 0);	//mark end of frame

  return t;
}

void render_audio(int t)
{
  SoundTIA::render(t);
}

void render_tile(unsigned char** raster, int tile, int xs, int ys)
{
  //int ptr = tile << 3;
	int ptr = tile << 6;

  for (int y = 0; y < 8; ++y)
  {
    int sy = ((ys + y) & 255);

    if (sy >= 8 && sy < 248)
    {
      sy-=8;
      //int pmask = 0x80;

      for (int x = 0; x < 8; ++x)
      {
        //int col = (vram[ptr + 0x0800] & pmask ? 1 : 0) | (vram[ptr + 0x1000] & pmask ? 2 : 0) | (vram[ptr + 0x1800] & pmask ? 4 : 0) | (vram[ptr + 0x2000] & pmask ? 8 : 0);
       // pmask >>= 1;
		  unsigned char col = vram_cache[ptr++];

        raster[sy][(xs + x) & 255] = vpal[col | bitplane_mask];
      }
    }

    //++ptr;
  }
}

void render_sprite(unsigned char** raster, int spr, int xs, int ys, bool flip_h, bool flip_v)
{
  int x_flip = 0;
  int y_flip = 0;

  if (flip_h) x_flip = 15;
  if (flip_v) y_flip = 15;

  int ptr = spr << 4;

  for (int y = 0; y < 16; ++y)
  {
    int sy = ((ys + (y ^ y_flip)) & 255);

    if (sy >= 8 && sy < 248)
    {
      sy-=8;

      for (int x = 0; x < 16; ++x)
      {
        int ptrx = ptr + ((x & 8) << 9);
        int mask = 128 >> (x & 7);

        unsigned char col = (vrom[ptrx + 0x0000] & mask ? 1 : 0) | (vrom[ptrx + 0x2000] & mask ? 2 : 0) | (vrom[ptrx + 0x4000] & mask ? 4 : 0) | (vrom[ptrx + 0x6000] & mask ? 8 : 0);

        if (col < 0x0f)
        {
          raster[sy][(xs + (x ^ x_flip)) & 255] = vpal[col];
        }
      }
    }

    ++ptr;
  }
}

void render_video(unsigned char** raster)
{
	if (vram_cache_update)
	{
		convert_tiles();
		vram_cache_update = false;
	}

  //render BG

  int vram_ptr = 0;

  if (vram_sel & 128) vram_ptr = 1024;

  int y = scr_yoff;

  for (int i = 0; i < 32; ++i)
  {
    int x = scr_xoff;

    for (int j = 0; j < 32; ++j)
    {
      render_tile(raster, vram[vram_ptr++], x, y);

      x += 8;
    }

    y += 8;
  }

  //render sprites

  for (int id = 0; id < 16; ++id)
  {
    if (!(spr_a[id] & SPR_ATTR_ENABLE)) continue;

    render_sprite(raster, spr_n[id], spr_x[id], spr_y[id], spr_a[id] & SPR_ATTR_FLIP_H ? true : false, spr_a[id] & SPR_ATTR_FLIP_V ? true : false);
  }
}

void pmem_load(const char* filename, unsigned short int adr)
{
  file_load(filename, &pmem[adr]);
}

void vrom_load(const char* filename, unsigned short int adr)
{
  file_load(filename, &vrom[adr]);
}

void init(bool _input_analog)
{
  pmem = global_ram1; //65536
  vrom = global_ram2; //32768
  vram = global_ram2 + 32768; //10240
  vram_cache = global_ram2 + 49152;	//16384

  input_analog = _input_analog;
  input_analog_value = 96 + 112 / 2;

  SoundTIA::init(0);

  vsync = 0;
  vram_sel = 0;
  vram_en = 0;
  bitplane_mask = 0;
  vram_cache_update = true;

  scr_xoff = 0;
  scr_yoff = 0;

  pio_d0 = 0;
  pio_d1 = 0;
  pio_d2 = 0;

  for (int i = 0; i < 16; ++i) spr_a[i] = 0;

  //make palette

  for (int i = 0; i < 256; ++i)
  {
    int ir = (i >> 3) & 7;
    int ig = i & 7;
    int ib = (i >> 6) & 3;

    double dr = 255.0 * r_v[ir] / r_v[0];
    double dg = 255.0 * g_v[ig] / g_v[0];
    double db = 255.0 * b_v[ib] / b_v[0];

    int r = (255 - (int)dr) & 255;
    int g = (255 - (int)dg) & 255;
    int b = (255 - (int)db) & 255;

#ifdef PALETTE_R8G8B8
    palette888[i] = (r << 16) | (g << 8) | b;
#endif
#ifdef PALETTE_R3G3B2
    palette332[i] = ((r >> 5) << 5) | ((g >> 5) << 2) | (b >> 6);
#endif
  }

  CPU8080::init(mem_rd, mem_wr, port_rd, port_wr);
}

void reset(void)
{
  CPU8080::reset();
  SoundTIA::reset();
}
}
