//S.A.M.E. (Soviet Arcade Machine Emulator) CC-BY shiru8bit (shiru@mail.ru)
//ESP32 port by the original author
//Original version has been made for Flash/AS3 back in 2016, later ported to Android/HAXE

//v1.01 04.07.23 completely different fast sound render, to avoid frameskip (low sound quality)
//v1.0  30.06.23 initial version (with frameskip and lower Foton sound pitch)

#include "SPIFFS.h"
#include <ESP_8_BIT_composite.h>
#include "soc/ledc_struct.h"
#include "driver/ledc.h"
#include "esp_wifi.h"
#include "esp_bt.h"

//ESP_8_BIT_composite videoOut(true);  //NTSC
ESP_8_BIT_composite videoOut(false);  //PAL

#define USE_DUAL_CORE

#define AUDIO_PIN   18  //can be any pin, uses LED PWM just like ESP_8_Bit

#define PIN_J_UP    32
#define PIN_J_DOWN  12
#define PIN_J_LEFT  4
#define PIN_J_RIGHT 0
#define PIN_J_A     26
#define PIN_J_B     19
#define PIN_J_X     23
#define PIN_J_Y     5

#define PAD_NONE    0x0000
#define PAD_UP      0x0001
#define PAD_DOWN    0x0002
#define PAD_LEFT    0x0004
#define PAD_RIGHT   0x0008
#define PAD_A       0x0010
#define PAD_B       0x0020
#define PAD_COIN    0x0040
#define PAD_TEST    0x0080

#define PALETTE_R3G3B2

void file_load(const char* filename, unsigned char* dst)
{
  char path[128];

  strcpy(path, "/");
  strcat(path, filename);

  File file = SPIFFS.open(path, "rb");

  if (!file) {
    Serial.print("There was an error opening ROM file ");
    Serial.println(path);
    return;
  }

  file.readBytes((char*)dst, file.size());
  file.close();

  Serial.print("Load OK ");
  Serial.println(path);
}

#define GLOBAL_RAM1_SIZE  65536
#define GLOBAL_RAM2_SIZE  65536

unsigned char* global_ram1;
unsigned char* global_ram2;

#include "EmuMain.h"



void joy_init(void)
{
  pinMode(PIN_J_UP, INPUT_PULLUP);
  pinMode(PIN_J_DOWN, INPUT_PULLUP);
  pinMode(PIN_J_LEFT, INPUT_PULLUP);
  pinMode(PIN_J_RIGHT, INPUT_PULLUP);
  pinMode(PIN_J_A, INPUT_PULLUP);
  pinMode(PIN_J_B, INPUT_PULLUP);
  pinMode(PIN_J_X, INPUT_PULLUP);
  pinMode(PIN_J_Y, INPUT_PULLUP);
}

unsigned int joy_read()
{
  unsigned int joy = 0;

  if (digitalRead(PIN_J_UP) == LOW) joy |= PAD_UP;
  if (digitalRead(PIN_J_DOWN) == LOW) joy |= PAD_DOWN;
  if (digitalRead(PIN_J_LEFT) == LOW) joy |= PAD_LEFT;
  if (digitalRead(PIN_J_RIGHT) == LOW) joy |= PAD_RIGHT;
  if (digitalRead(PIN_J_A) == LOW) joy |= PAD_A;
  if (digitalRead(PIN_J_B) == LOW) joy |= PAD_B;
  if (digitalRead(PIN_J_X) == LOW) joy |= PAD_COIN;
  if (digitalRead(PIN_J_Y) == LOW) joy |= PAD_TEST;

  return joy;
}


void IRAM_ATTR sound_timer_ISR()
{
  //playing back audio via LED PWM, just like in the ESP_8_Bit

  auto& reg = LEDC.channel_group[0].channel[0];
  reg.duty.duty = audio_queue_get() << 4; // 25 bit (21.4)
  reg.conf0.sig_out_en = 1;
  reg.conf1.duty_start = 1;
  reg.conf0.clk_en = 1;
}



void clear_frame_buffer(void)
{
  for (int page = 0; page < 2; ++page) //video is double buffered, clear both buffer
  {
    uint8_t** frameBufferLines = videoOut.getFrameBufferLines();

    for (int i = 0; i < 240; ++i) memset(frameBufferLines[i], 0, 256);

    videoOut.waitForFrame();
  }
}



int game_id = 0;

void main_loop(void *parameter)
{
  hw_timer_t *sound_timer = timerBegin(0, 1, true);
  timerAttachInterrupt(sound_timer, &sound_timer_ISR, true);
  timerAlarmWrite(sound_timer, 80000000 / SAMPLE_RATE, true);
  timerAlarmEnable(sound_timer);

  videoOut.begin();

  while (1)
  {
    clear_frame_buffer();
    audio_queue_init();

    memset(global_ram1, 0, GLOBAL_RAM1_SIZE);
    memset(global_ram2, 0, GLOBAL_RAM2_SIZE);

    load_game(game_id);

    emulate(0, 1); //skip some game initialization frames

    while (1)
    {
      int joy = joy_read();

      int t = emulate(joy, 1);

      render_audio(t);

      videoOut.waitForFrame();

      uint8_t** frameBufferLines = videoOut.getFrameBufferLines();  //the video library implements double buffering

      render_video(frameBufferLines);

      if (joy & PAD_TEST) //end current game
      {
        while (1) //wait for button release, to avoid skipping games
        {
          joy = joy_read();
          if (!(joy & PAD_TEST)) break;
          delay(1);
        }
        break;
      }
    }

    ++game_id;   //next game

    if (game_id >= GAMES_ALL) game_id = 0;
  }
}



void setup()
{
  rtc_clk_cpu_freq_set(RTC_CPU_FREQ_240M);

  global_ram1 = (unsigned char*)heap_caps_malloc(GLOBAL_RAM1_SIZE, MALLOC_CAP_8BIT);
  if (!global_ram1) Serial.println("global_ram1 null");

  global_ram2 = (unsigned char*)heap_caps_malloc(GLOBAL_RAM2_SIZE, MALLOC_CAP_8BIT);
  if (!global_ram2) Serial.println("global_ram1 null");

  if (!SPIFFS.begin(true))
  {
    Serial.println("An Error has occurred while mounting SPIFFS");
    return;
  }

  joy_init();

  ledcSetup(0, 2000000, 7);  // 625000 khz is as fast as we go w 7 bits
  ledcAttachPin(AUDIO_PIN, 0);
  ledcWrite(0, 0);

#ifdef USE_DUAL_CORE
  xTaskCreatePinnedToCore(main_loop, "Core 0 task", 1024 * 3, NULL, 0, NULL, 0);
#else
  main_loop(0);
#endif
}



void loop()
{
}
